/* 
 * 
 * Copyright (c) Year(s), 2013, Luis M. S. Russo and Alexandre 
 * P. Francisco / KDBIO / INESC-ID, <qhv@kdbio.inesc-id.pt> 
 *  
 * Any published media that is related with to use of the distributed 
 * software, or derived software, must contain a reference to "Quick
 * HyperVolume, Luís M. S. Russo, Alexandre P. Francisco IEEE Trans. 
 * Evolutionary Computation 18(4): 481-502(2014)".
 *  
 * Permission to use, copy, modify, and/or distribute this software for 
 * any purpose with or without fee is hereby granted, provided that the 
 * above copyright notice and this permission notice appear in all 
 * copies. 
 *  
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL 
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE 
 * AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL 
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR 
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER 
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR 
 * PERFORMANCE OF THIS SOFTWARE. 
 * 
 */


#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <strings.h>
#include "quickhvolume.h"
#include "constants.h"
#include "HSO.h"
#include "subsets.h"
#include "division.h"
#include "splitter.h"
#include "macros.h"
#include "inexclusion.h"

/* global variables */

/* file scope macro definitions */

/* file scope type declarations */

/* file scope variables */

#ifdef COUNT                    /**< Count QuickHvolumeR T(n) */
static double tops;             /**< Count the number of ops, instead of time */
#endif  /* COUNT */
	
/* file scope functions declarations */

/** 
 * The recursive version is private for this file
 * 
 * @param zero The min coordinates point
 * @param one The max coordinates point
 * @param n Number of selected points
 * @param idx List of indexes
 * @param PS List of points
 * 
 * @return The resulting hyperVolume
 */
static double quickHVolumeR(point* zero,
                            point* one,
                            int n,
                            int* idx,
                            point* PS);

#ifdef COUNT
static double WrappedQuickHVolumeR(point* z,
                                   point* o,
                                   int n,
                                   int* idx,
                                   point* PS);
#endif /* COUNT */

/** 
 * Drops the element at position j, by moving to the end of the array
 * and decreasing n
 * 
 * @param idx The index array
 * @param n Last position in array
 * @param j The index to drop
 */
static void drop(int* idx, int* n, int j);

/* file scope functions */

static void drop(int* idx, int* n, int j)
{
  (*n)--;
  idx[j] = idx[*n];
}

/* paper: 1 */
static double quickHVolumeR(point* z,
                            point* o,
                            int n,
                            int* idx,
                            point* PS)
/* paper: 0 */
#ifdef COUNT
{
  tops += n;
  WrappedQuickHVolumeR(z, o, n, idx, PS);

  return 0;
}

static double WrappedQuickHVolumeR(point* z,
                                  point* o,
                                  int n,
                                  int* idx,
                                  point* PS)
#endif /* COUNT */
/* paper: 1 */
{
  double hv;                    /**< hyperVolume */
  division d;                   /**< The division structure */
  int i, j;                     /**< Counters */
  point p;                      /**< Pivot */
  point pp;                     /**< Temporary point */
/* paper: 0 */
  point ppp;                    /**< Another temporary point */
/* paper: 1 */
  point newo;                   /**< The new o point */
  point newz;                   /**< The new z point */
  double max;                   /**< maximize objective */

  max = -DBL_MAX;               /**< Forces initial loop */

  ///1. Select Pivot
  i = 0;
  while(i < n)
    {
      intercept(&pp, o, &(PS[idx[i]]));
      hv = objective(z, &pp, o);
      if(hv > max)
        {                       /**< New Pivot Candidate */
          max = hv;
          j = i;
        }
      i++;
    }

  /// Separate Pivot
  intercept(&p, o, &(PS[idx[j]]));
  hv = HV(z, &p);
  drop(idx, &n, j);

  /// 2. Divide Points
  i = 0;
  while(i < n)
    {
      intercept(&pp, o, &(PS[idx[i]]));
      push(idx[i], classify(&p, &pp));
      i++;
    }
  d = newDivision(PS, o, &p);

  /// 3. Conquer
  while(hasNext(d))
    {
      idx = next(d, &j, &n);
      projectZero(&newz, &p, j, z);
      projectOne(&newo, &p, j, o);

/* paper: 0 */
      if(n < N)
        {  /**< For very few points use cache oblivious */
          switch(n)
            {
            case 1:             /**< Simple volume */
              intercept(&pp, &newo, &(PS[idx[0]]));
              hv += HV(&newz, &pp);
              break;
            case 2:             /**< Simple IEX */
              intercept(&pp, &newo, &(PS[idx[0]]));
              hv += HV(&newz, &pp);
              intercept(&ppp, &newo, &(PS[idx[1]]));
              hv += HV(&newz, &ppp);
              intercept(&pp, &pp, &ppp); /**< Exclude part */
              hv -= HV(&newz, &pp);
              break;
            default:
              if(HSOL < n)
                hv += HSO(&newz, &newo, n, idx, PS);
              else
                hv += InExClusion(&newz, &newo, n, idx, PS);
              break;
            }
        }
      else /**< Use recursive call */
/* paper: 1 */
        {
          /* if(on > 10 && bits(j)+log2((double)n/(double)on) > 3) */
          /*   printf("bits:%d  number:%d  fraction:%f\n", bits(j), on, log2((double)n/(double)on)); */
          hv += quickHVolumeR(&newz, &newo, n, idx, PS);
        }
    }
  freeDivision(&d);

  return hv;
}
/* paper: 0 */

/* public functions */

double quickHVolume(int n, point* PS)
{
  double res;
  int* idx;
  int i;

#ifdef COUNT
  tops = 0;
#endif /* COUNT */

  idx = (int*) malloc(n*sizeof(int));
  i = 0;
  while(i < n)
    {
      idx[i] = i;
      i++;
    }

  res = quickHVolumeR(&cZero, &cOne, n, idx, PS);
  
  free(idx);

#ifdef COUNT
  printf("%d %d %f\n", n, D, tops);
#endif /* COUNT */

  return res;
}

/* ------------------------------------------------------ */
